--[[
Copyright (c) 2017 raidho36/rcoaxil

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
]]--

-- JIT version requires LuaJIT running
assert ( type ( jit ) == 'table' and jit.status ( ) == true )

local msqrt, msin, mcos, matan2 = math.sqrt, math.sin, math.cos, math.atan2

local Vec3 = setmetatable ( { }, { __call = function ( class, ... ) return class.new ( ... ) end } )
Vec3.__index = Vec3
Vec3.__call = function ( self, other )
	if other then
		self.x = other.x
		self.y = other.y
		self.z = other.z
		return self
	else
		return Vec3.new ( self.x, self.y, self.z )
	end
end

function Vec3.new ( x, y, z )
	local self = setmetatable ( { }, Vec3 )
	self.x = x or 0
	self.y = y or 0
	self.z = z or 0
	return self
end

function Vec3:clone ( )
	return Vec3.new ( self.x, self.y, self.z )
end

function Vec3:copy ( v )
	self.x = v.x
	self.y = v.y
	self.z = v.z
	return self
end

--from spherical coordinates (azimuth, elevation, length)
function Vec3:spherical ( a, e, l )
	local acl = mcos ( e ) * l
	self.x = mcos ( a ) * acl
	self.z = msin ( a ) * acl
	self.y = msin ( e ) * l
	return self
end

function Vec3:add ( v )
	self.x = self.x + v.x
	self.y = self.y + v.y
	self.z = self.z + v.z
	return self
end

function Vec3:sub ( v )
	self.x = self.x - v.x
	self.y = self.y - v.y
	self.z = self.z - v.z
	return self
end

function Vec3:mul ( v )
	self.x = self.x * v.x
	self.y = self.y * v.y
	self.z = self.z * v.z
	return self
end

function Vec3:div ( v )
	self.x = self.x / v.x
	self.y = self.y / v.y
	self.z = self.z / v.z
	return self
end

function Vec3:sum ( a, b )
	self.x = a.x + b.x
	self.y = a.y + b.y
	self.z = a.z + b.z
	return self
end

function Vec3:diff ( a, b )
	self.x = a.x - b.x
	self.y = a.y - b.y
	self.z = a.z - b.z
	return self
end

function Vec3:prod ( a, b )
	self.x = a.x * b.x
	self.y = a.y * b.y
	self.z = a.z * b.z
	return self
end

function Vec3:quot ( a, b )
	self.x = a.x / b.x
	self.y = a.y / b.y
	self.z = a.z / b.z
	return self
end

function Vec3:scale ( f )
	self.x = self.x * f
	self.y = self.y * f
	self.z = self.z * f
	return self
end

function Vec3:invert ( )
	self.x, self.y, self.z = -self.x, -self.y, -self.z
	return self
end

function Vec3:normalize ( )
	local l = 1 / msqrt ( self.x * self.x + self.y * self.y + self.z * self.z )
	self.x = self.x * l
	self.y = self.y * l
	self.z = self.z * l
	return self
end

function Vec3:resize ( l )
	l = l / msqrt ( self.x * self.x + self.y * self.y + self.z * self.z )
	self.x = self.x * l
	self.y = self.y * l
	self.z = self.z * l
	return self
end

function Vec3:len ( )
	return msqrt ( self.x * self.x + self.y * self.y + self.z * self.z )
end

function Vec3:dir ( )
	return matan2 ( self.z, self.x ), matan2 ( self.y, msqrt ( self.x * self.x + self.z * self.z ) )
end

function Vec3:dot ( v )
	return self.x * v.x + self.y * v.y + self.z * v.z
end

--interpolate to position between A and B
function Vec3:lerp ( a, b, t )
	local i = 1 - t
	self.x = a.x * i + b.x * t
	self.y = a.y * i + b.y * t
	self.z = a.z * i + b.z * t
	return self
end

--distance to another vec3
function Vec3:dist ( v )
	return msqrt ( ( self.x - v.x ) ^ 2 + ( self.y - v.y ) ^ 2 + ( self.z - v.z ) ^ 2 )
end

--spherical angles to another vec3
function Vec3:angle ( v )
	local xx, zz = v.x - self.x, v.z - self.z
	return matan2 ( zz, xx ), matan2 ( v.y - self.y, msqrt ( xx * xx + zz * zz ) )
end

function Vec3:pack ( )
	return { self.x, self.y, self.z }
end

function Vec3:unpack ( )
	return self.x, self.y, self.z
end

------------------------------
-- additional functionality --
------------------------------

--sets to cross product between vec3 A and B
function Vec3:cross ( a, b )
	self.x, self.y, self.z = a.y * b.z - b.y * a.z, a.z * b.x - b.z * a.x, a.x * b.y - b.x * a.y
	return self
end

-- intersects ray P,V with plane O·N
-- ray direction criterion is t >= 0
function Vec3:ray2plane ( p, v, o, n ) --a, x
	-- Ray: P + V * t
	-- Plane: ( P - O ) · N = 0
	-- Ray × Plane = ( P + V * t - O ) · N = 0
	-- V * t · N = -( P - O ) · N
	-- t = ( O - P ) · N / V · N
	local t = ( ( o.x - p.x ) * n.x + ( o.y - p.y ) * n.y + ( o.z - p.z ) * n.z ) / ( v.x * n.x + v.y * n.y + v.z * n.z )
	self.x = p.x + v.x * t
	self.y = p.y + v.y * t
	self.z = p.z + v.z * t
	return self, t
end

-- intersects ray A,X with triangle J,K,L
-- because it uses barycentric coordinates, works fine for paralellograms too
-- ray direction criterion is t >= 0
-- intersection criterion for triangle is v >= 0; u >= 0; v + u <= 1
-- intersection criterion for square is 1 >= v >= 0; 1 >= u >= 0
function Vec3:ray2triangle ( p, v, j, k, l )
	-- Ray: P + V * t
	-- Triangle: ( 1 - u - v ) * J + u * K + v * L = J + u * JK + v * JL
	-- Ray × Triangle: J + JK * u + JL * v = P + V * t
	-- J - P = JK * u + JL * v - V * t
	-- Cramer's rule: NxN linear equations matrix with N variables can be
	-- solved as xi = d(Ai) / d(A) where Ai is matrix with i'th column
	-- substituted by solution column-vector
	-- [ t ]         1          [ | ( J - P ), JK, JL | ]
	-- [ u ] = -------------- * [ | -V, ( J - P ), JL | ]
	-- [ v ]   | -V, JK, JL |   [ | -V, JK, ( J - P ) | ]
	-- Determinant of square A,B,C matrix is triple product A · ( B × C )
	-- Swapping any two components negates the product; denoting J - P as T,
	-- [ t ]         1          [ T × JK · JL ]
	-- [ u ] = -------------- * [ V × JL · T  ]
	-- [ v ]    V × JL · JK     [ T × JK · V  ]
	local jlx, jly, jlz = l.x - j.x, l.y - j.y, l.z - j.z -- JL = L - J
	local vjlx, vjly, vjlz = v.y * jlz - jly * v.z, v.z * jlx - jlz * v.x, v.x * jly - jlx * v.y -- VJL = V × JL
	local jkx, jky, jkz = k.x - j.x, k.y - j.y, k.z - j.z -- JK = K - J
	local det = 1 / ( jkx * vjlx + jky * vjly + jkz * vjlz ) -- d = 1 / VJL · JK

	self.x, self.y, self.z = p.x - j.x, p.y - j.y, p.z - j.z -- T = P - J
	local u = ( vjlx * self.x + vjly * self.y + vjlz * self.z ) * det -- u = VJL · T * d

	vjlx, vjly, vjlz = self.y * jkz - jky * self.z, self.z * jkx - jkz * self.x, self.x * jky - jkx * self.y -- TJK = T × JK
	local v = ( vjlx * v.x + vjly * v.y + vjlz * v.z ) * det -- v = TJK · V * d
	local t = ( vjlx * jlx + vjly * jly + vjlz * jlz ) * det -- t = TJK · JL * d

	self.x = p.x + v.x * t -- X = P + V * t
	self.y = p.y + v.y * t
	self.z = p.z + v.z * t
	return self, t, u, v
end

------------------------------
--    function  variants    --
------------------------------

--length squared
function Vec3:len2 ( )
	return self.x * self.x + self.y * self.y + self.z * self.z
end

--distance squared
function Vec3:dist2 ( v )
	return ( self.x - v.x ) ^ 2 + ( self.y - v.y ) ^ 2 + ( self.z - v.z ) ^ 2
end

local ffi = require ( "ffi" )

ffi.cdef ( "typedef struct rsml_vec3 { double x, y, z; } rsml_vec3;" )
Vec3.new = ffi.typeof ( "rsml_vec3" )
ffi.metatype ( "rsml_vec3", Vec3 )

return Vec3
